#!/bin/sh
#
# Copyright (c) 1999-2001, ``vman'' (moopowah@yahoo.com)
# All rights reserved.  Please refer to the LICENSE file
# included in vCheck's package for more information, or
# the homepage http://vware.hypermart.net.
#
#  Compatible with glFtpD-1.22b and OpenFtpD 1.0.0 (Linux, FreeBSD)
#        (do you have a Solaris box I can use to port to?)
#
#  vCheck revision: 0.1.0
#
#
###-[   Configuration   ]-######################################################
#
###[ What daemon? ]
#What daemon are you using?
#-> "1" (glFtpD, www.glftpd.org)
#-> "2" (OpenFtpD, www.openftpd.org) (this is still developmental and not tested 
#fully)
myDaemon="1"


#Where is your bin directory (where you would keep checksum binary in)?
#-> "/path/to/the/dir"
binDir="/bin"

#Where is your group file (this currently only matters for glftpd users)?
groupFile="/etc/group"


###[ SFV ]
#Perform SFV for sfv-based files with the $binDir/asfv utility (this is a 
#modified aSFV with permission of the author and should not be used/distributed 
#as the official aSFV program).
#-> "/some/tree/*:/could/have/another/path:/or/another/????:/site/*/cows"
#-> ""
sfvcheck=""

#If $sfvcheck is disabled for a particular directory, you may fall back on this 
#method of file integrity checking.  Of course for glftpd, calc_crc must be 
#enabled for this directory/file.  Thus, you can potentially have both checksum 
#and on-the-fly sfv enabled for different parts of your site.  If a file does 
#not match $sfvcheck nor $flysfv, then it will be accepted without check.  
#Observe the different forms that you can use (even negation!).
#-> "/some/*/cows2:/some/t2/*:/site/0day/????:/site/[bB][lL][aA]/*:!/site/bla*"
#-> ""
flysfv="*"

#Allow sfv file to *not* be present and still accept files?  Once the sfv file 
#is uploaded completion will take place, along with testing each file; should 
#any files fail crc, the user's who own it will decrement in their stats and 
#credits appropriately.  *THIS DOES NOT WORK YET*
#-> "/site/prewhores/*"
#-> ""
noSfv=""



###[ ZIP ]
#Perform an integrity check for zip files (must have unzip binary)?
#-> "/site/0DAY/*:/site/private/0DAY/*:/site/public/0DAY2/*"
#-> ""
unzip="*"


#Extract the contents of file_id.diz from zip files and place it in .message?
#-> "/site/0DAY/*:/site/private/0DAY/*:/site/public/0DAY2/*"
#-> ""
zipDiz="*"

#Which dir is actually made for zip files?  Some iso titles have zip file 
#accompaniment and vCheck can't decide for you (what if the zip is uploaded 
#first for an iso release?).  So explicitly define where zip-completion gets 
#priority.
#-> "/site/0day/*"
#-> ""
zipRoutine="/site/0day/*"

#Add a `comment' to zip files that are uploaded (the little banner that scrolls 
#as you unzip)?
#-> "/path/to/your/comment.file"
#-> ""
zipComment=""

#Add a file to zip files that are uploaded (typically an nfo file).
#-> "/path/to/your/site.nfo"
#-> ""
zipNfo=""


###[ MP3 ]
#Extract mp3 id3/headers?
#-> "/site/MP3/*:/site/public/MP3/*"
#-> ""
mp3Info="*"

#If the above is enabled, would you like a dir made in the release's with some 
#of the info when the release is complete (such as bitrate, genre, etc.).
#-> "/site/MP3/*:/site/public/MP3/*"
#-> ""
mp3InfoDir="*"

#If $mp3Info is enabled, would you also like symlinks made based on genre?  An 
#example would be /site/MP3/Mp3sByGenre/Rap/MyRapSong as a symlink to where ever 
#the actual release resides.  There is another script which will delete symlinks 
#that are no longer in use.  If the genre dir does not exist, it will 
#automatically be made.
#-> "/site/MP3/Mp3sByGenre"
#-> ""
mp3GenreSort="/site/mp3-by-genre"

#Same as above but sort based on dir name (not the id3 title, but dir name).
#-> "/site/MP3/Mp3sByAlphaNumeric"
#-> ""
mp3TitleSort="/site/mp3-by-title"

#Same as above but sort based on group (*-GRP). 
#-> "/site/MP3/Mp3sByGroup"
#-> ""
mp3GroupSort="/site/mp3-by-group"


###[ General/Misc ]
#
#Should vCheck cache the .racestart file in the release's dir even after the 
#race is finished?  This file distinguishes a release in the database and once 
#it is deleted and a file is uploaded (let's say in the middle of the race), 
#vCheck will think it's an entirely *new* release you have just begun.  So if a 
#race is done, the file is safe to be deleted (makes it cleaner); however, if 
#you need to reupload one file after a release is done for some reason, it'll 
#think it is a #new release with only one file uploaded.
#-> "/site/iso/*"
#-> ""
cacheRacestart="*"

#Announce the amount/estimated size of the expected files (iso/mp3/zip)?
announceExpectedFiles="*"

#Create a .message file?
#-> "/site/public/*:/site/public2/*"
#-> ""
msgFile="*"

#The amount of disks a release must have in order to proceed with the 
#race/completion portions (meaning exit after file integrity check).
#-> <positive_integer>
#-> 0
requiredDisks="0"

#Perform as an integrity checker *only*; do not advance to any logging- or 
#database-related procedures (no completion, race, etc.).  Note that $zipDiz, 
#$zipComment, and $zipNfo are performed before the exit.
#-> "/site/private/*:/site/admins/*"
#-> ""
denyPaths="/site/*/_Groups/*"

#This enables a progress directory to created, which points out how many files 
#are missing while a race is in session, and ultimately creates the directory 
#specified here in the release directory when the race has been finished.
#-> "this release is complete"
#-> ""
makeCompletionDir="[_COMPLETE_]"

#vCheck *requires* a mySQL server (www.mysql.com) to be installed; specify the 
#login information here (use IP, not hostname).
#-> "<ip> <login> <password>"
database="127.0.0.1 vcheck kcehcv99"

#Specify a directory, if any, to create symlinks to incomplete releases
#-> "/site/incompltes"
#-> ""
incLink="/site/INCOMPLETES"

#Create FILENAME-missing files (0-byte) until file is uploaded (then it's 
#removed).  This was requested by several people, and I've seen countless others 
#use it, so here it is (remember, for: rar, mp3, zip).  If $createMissing is 
#enabled, then tell vCheck what string to append to the 0byte filename.
createMissing="*"
createMissingString="-missing"

#Send data to logs when a new racer (user, and group) is found?  Since this 
#sends a good deal of data to the log, don't leave it enabled if you don't plan 
#on using it. 
#-> "/site/iso/*:/site/0day/*"
#-> ""
announceNewRacer="*"
#
################################################################################
#
#send the variables to the program for processing :)
exec /bin/tclsh /bin/vcheck/vcheck.tcl "$1" "$2" "$3" \
"$cacheRacestart" "$binDir" "$groupFile" \
"$sfvcheck" "$flysfv" "$incLink" "$announceExpectedFiles" \
"$unzip" "$zipRoutine" "$noSfv" "$zipDiz" "$zipComment" "$zipNfo" \
"$mp3Info" \
"$msgFile" "$requiredDisks" "$denyPaths" "$makeCompletionDir" "$myDaemon" \
"$database" "$mp3InfoDir" "$mp3GenreSort" "$mp3TitleSort" "$mp3GroupSort" \
"$createMissing" "$announceNewRacer" "$createMissingString" \
2>> /tmp/vcheckErrors.log


#Want to debug?
#per default, a logfile of errors is made and appended to.  You may also attempt 
#to debug errors by chrooting to /glftpd, creating the SPEED and GROUP 
#environment variables, and running vcheck.sh by hand (don't forget to 
#discontinue the redirection of the error stream (the above `2>>').


