#!/bin/tclsh
#
# vCheck (vman's checker) v0.1.0
# advanced post-upload file processor/verifier for glFtpD and OpenFTPD
#
# official homepage      : http://vware.hypermart.net
# author's email         : moopowah@yahoo.com
# author's efnet nick    : v-man
#
# This software is OSI Certified Open Source Software under The BSD License.
# (For more information, please see: http://www.opensource.org)
#
################################################################################
#
# Copyright (c) 1999-2001, ``vman'' (moopowah@yahoo.com)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following four (4) conditions
# are met:
#
#      1. Redistributions of source code must retain the above copyright notice,
#         this list of conditions and the following disclaimer.
#
#      2. Redistributions in binary form must reproduce the above copyright
#         notice, this list of conditions and the following disclaimer in the
#         documentation and/or other materials provided with the distribution.
#
#      3. You, the licensee, may not charge a fee whatsoever for the
#         redistribution of this software in any format without specific prior
#         written permission and arrangements from the original author and
#         copyright holder, ``vman'' (moopowah@yahoo.com).
#
#      4. Neither the author, ``vman'' (moopowah@yahoo.com) not his
#         contributors may be used to endorse or promote products derived from
#         this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
################################################################################




#TODO (sloppy notes by vman)
#1: use vcheck:stringToList for warning/failure/error msgs to user
#2: pull out zip dates from diz (rename $extras to $someDateName... since both sfv and zip will now have it for [race]).
   #DATES
   #13/01/01, <01/13/2001>, 13.01.01,, 01/13/2001, [DATE:01/13/01], _01/13/2001_, [01.13.01]
   #_[jAN 2001], [13-01-2001], -{13.01.2001}-, 13/o1/2oo1, [ 01.13.2oo1 ], [13 January 01]
   #05-01-2001, [Date:04/01/01], January 5th/2001, 05-Jan-2001, o5-o1-o1, (01-05) (assume the year?! :P)
   #[jAN 2001], [DATE:01-04-2001], Jan 11th 2000
   #note: euro is dd/mm/yy[yy]
   #note: usa is mm/dd/yy[yy]
#3: add diskcounts: " 01/01 ", [Disks: 03/03], [ DISK: 01x2MB ], 250- 01/18 (^...)
#4: sort by dirname/grp for *any* type of file?  as long as $showCompletion is 1 and it's "done" (just use mp3 code in new place).
#5: add in 'sfv any time' support -- shouldn't be hard, much like the $newDiz (cache crc's, if calc_crc is on, to a file).
#6: get max/low speeds
#7: make !racers <dir>  to tell u how many racers are going for it + how many files left.  post rid with the release like #234 so they query that!
#8: this release got an avg bandwidth of 23433kB/s'

#fixed
#1: the mp3Info wasn't being sent to the bot correctly (elements are in braces now)


#########
# Start #
################################################################################
proc vcheck:checkForOk {} { global currFile
	set newCurrFile ""
	# Examples of filenames: myfile.rar.ok, myfile.ok.rar
	foreach fileNamePortion [split $currFile "."] {
		if {$fileNamePortion != "ok"} {
			append newCurrFile "$fileNamePortion\."
		}
	}
	# Take out appended period.
	set newCurrFile [string range $newCurrFile 0 end-1]
	if {$currFile != $newCurrFile} {
		file rename -force -- $currFile $newCurrFile
		set currFile $newCurrFile
	}
	return 0
}


proc vcheck:db:query {myQuery} { global mysql
	# Take in the query line, query the db, and create the list $rows to the caller's proc level.
	sql query $mysql(conn) $myQuery
	upvar 1 rows rows
	set rows ""
	while {[set row [sql fetchrow $mysql(conn)]] != ""} {
		if {[llength $row] != 1} {
			lappend rows $row
		} else {
			lappend rows [lindex $row 0]
		}
	}
	catch {sql endquery $mysql(conn)}
	if {([string index $rows 0] == "\{") && ([llength $rows] == 1)} {
		set rows [lindex $rows 0]
	}
	return 0
}


proc vcheck:db:getReleaseId {} { global mysql rid currPath currTime timeStampFile
	set timeStamp ""
	if {![file exist "$currPath/$timeStampFile"]} {
		# This release has *not* been entered in the database at all.  Let's cache the time so we can query it later.  Note, everything is based on $currTime which is global (the first file's timeStamp, probably this file being uploaded, will have this timestamp as well).
		set timeStamp [clock format $currTime -format "%Y-%m-%d %H:%M:%S"]
		set fout [open "$currPath/$timeStampFile" w]
		puts $fout $timeStamp
		close $fout
		sql exec $mysql(conn) "INSERT INTO Releases VALUES (NULL, '$currPath', '$timeStamp');"
	} else {
		# The cached timestamp exists, so let's ring it up.
		set fin [open "$currPath/$timeStampFile" r]
		set timeStamp [gets $fin]
		close $fin
	}
	vcheck:db:query "SELECT releaseId FROM Releases WHERE fullDir='$currPath' AND timeStamp='$timeStamp';"
	# We are assured that $rows has the releaseId for us now (whether the record existed initially or not).
	set rid [lindex $rows 0]
	return 0
}


proc vcheck:db:connect {} { global argv mysql revision
	# Login to mysql server and switch to the database.
	if {[catch {load /usr/lib/sql.so} err]} {
		puts $err
		vcheck:db:safeExit 2
	}

	# The below copyright must stay by license and simple morality (have some integrity for goodness sakes!).
	puts "\nvCheck v$revision   -   Copyright (c) 1999-2001, vman (moopowah@yahoo.com).  All rights reserved.\n"

	# Login with <ip> <user> <pass>.
	if {[catch {set mysql(conn) [sql connect [lindex [lindex $argv 22] 0] [lindex [lindex $argv 22] 1] [lindex [lindex $argv 22] 2]]} err]} {
		puts "Error: Opening database connection: $err"
		vcheck:db:safeExit 2
	}
	sql selectdb $mysql(conn) $mysql(db)
	return 0
}


proc vcheck:db:delRecord {myTable myFile} { global mysql rid
	sql exec $mysql(conn) "DELETE FROM $myTable WHERE releaseId=$rid AND fileName='$myFile';"
	return 0
}


proc vcheck:completion:getCompletedFiles {} { global mysql rid currPath currTime currFile argv
	#1: flush duplicates
	#2: verify the completed files *do* in fact reside on the drive (i.e., if someone deleted a file, then delete that record from the database)
	#3: return the completed files

	# We know one file is complete: our current file.
	set completedFiles $currFile
	# Let's set $rid for all further database commands relating to this particular release dir.
	vcheck:db:getReleaseId
	# Flush the $currFile if it existed in the database already (in case someone must have deleted it and is now re-uploading)
	vcheck:db:delRecord ReleasesDetails $currFile
	# Get the other completed files from the database.
	vcheck:db:query "SELECT fileName FROM ReleasesDetails WHERE releaseId=$rid;"
	set createZeroByteMissing [vcheck:matchPath [lindex $argv 27]]
	foreach myFile $rows {
		if {[file exist "$currPath/$myFile"]} {
			lappend completedFiles $myFile
		} else {
			# It seems someone has deleted a file from the drive that is still recorded in the database.  Clean the record.
			vcheck:db:delRecord ReleasesDetails $myFile
			# Now add it back to IncompleteReleases table.
			sql exec $mysql(conn) "INSERT INTO IncompleteReleases VALUES ($rid, '$myFile');"
			if {$createZeroByteMissing} {
				set fout [open "$myFile[lindex $argv 29]" w]
				close $fout
			}
		}
	}
	return [lsort -dictionary -increasing $completedFiles]
}


proc vcheck:db:safeExit {exitStatus} { global mysql tcl_platform currFile
	# FreeBSD dies when trying to end the session with mysqld (probably because the tcl-sql is for an older version and we're forcing it on newer msyql daemons).
	catch {sql endquery $mysql(conn)}
	if {($tcl_platform(os) != "FreeBSD") && ($mysql(conn) != "") && ([catch {sql disconnect $mysql(conn)} err])} {
		puts "Failure: Database disconnection: $err"
	}
	if {$exitStatus != 0} {
		catch {file delete -force -- $currFile}
	}
	# Goodbye... :-)
	exit $exitStatus
}


proc vcheck:matchPath {dirs} { global currPath
	foreach myDir [split $dirs ":"] {
		if {[string index $myDir 0] == "!"} {
			if {![string match [string range $myDir 1 end] $currPath]} {
				return 1
			}
		} else {
			if {[string match $myDir $currPath]} {
				return 1
			}
		}
	}
	return 0
}


proc vcheck:validateSfv {myFile} { global sfvMaxSize
	if {[file size $myFile] >= $sfvMaxSize} {
		puts "failure: opening sfv file: file \"$myFile\" ([file size $my_file] bytes) is too large to parse."
		vcheck:db:safeExit 1
	}
	return 0
}


proc vcheck:mp3:findBitrate {mpeg layer value1 value2 value3 value4} {
	if {$layer == 1} {
		return $value1
	} elseif {$layer == 2} {
		return $value2
	} elseif {$mpeg == 1} {
		return $value3
	}
	return $value4
}


proc vcheck:filterString {myString} {
	# Filter out some characters that mysql does not like.
	regsub -all -- {'} $myString "`" myString
	regsub -all -- {\\} $myString "`" myString
	regsub -all -- {\{} $myString "<" myString
	regsub -all -- {\}} $myString ">" myString
	# Take out all control characters.  After the comment/album, etc. for id3 tag the hex character 0x00 starts and is repeated until 0x0A (marking the end).  Tcl recognizes these as control characters.  Note the double brackets, `[:cntrl:]' is actually a list of characters, thus it needs to be enclosed with [ and ].
	regsub -all -- {[[:cntrl:]]} $myString {} myString
	return $myString
}


proc vcheck:binaryToDecimal {myNum} {
	# Must use double percision so the numbers do not wrap.
	set posVal 1.0
	set myResult 0.0
	set totalDigits [string length $myNum]
	for {set a 0} {$a < $totalDigits} {incr a} {
		set myResult [expr $myResult + ([string index $myNum end-$a] * $posVal)]
		set posVal [expr $posVal * 2.0]
	}
	# Don't return the double percision portion (".#"), just a clean integer.
	return [string range $myResult 0 end-2]
}


proc vcheck:mp3:getId3 {} { global currFile
	# I grabbed these genre values from: http://www.dv.co.yu/mpgscript/mpeghdr.htm
	# The rest came from tag.cxx from the mp3info program everyone but me uses. ;)
	array set mp3Genre "0 Blues 1 {Classic Rock} 2 Country 3 Dance 4 Disco 5 Funk 6 \
			Grunge 7 Hip-Hop 8 Jazz 9 Metal 10 {New Age} 11 Oldies 12 Other 13 Pop 14 R&B \
			15 Rap 16 Reggae 17 Rock 18 Techno 19 Industrial 20 Alternative 21 Ska \
			22 {Death Metal} 23 Pranks 24 Soundtrack 25 Euro-Techno 26 Ambient 27 Trip-Hop \
			28 Vocal 29 Jazz+Funk 30 Fusion 31 Trance 32 Classical 33 Instrumental 34 Acid \
			35 House 36 Game 37 {Sound Clip} 38 Gospel 39 Noise 40 AlternRock 41 Bass \
			42 Soul 43 Punk 44 Space 45 Meditative 46 {Instrumental Pop} 47 {Instrumental Rock} \
			48 Ethnic 49 Gothic 50 Darkwave 51 Techno-Industrial 52 Electronic 53 Pop-Folk \
			54 Eurodance 55 Dream 56 {Southern Rock} 57 Comedy 58 Cult 59 Gangsta 60 {Top 40} \
			61 {Christian Rap} 62 Pop/Funk 63 Jungle 64 {Native American} 65 Cabaret 66 {New Wave} \
			67 Psychadelic 68 Rave 69 Showtunes 70 Trailer 71 Lo-Fi 72 Tribal 73 {Acid Punk} \
			74 {Acid Jazz} 75 Polka 76 Retro 77 Musical 78 {Rock & Roll} 79 {Hard Rock} \
			80 Folk 81 Folk-Rock 82 {National Folk} 83 Swing 84 {Fast Fusion} 85 Bebob \
			86 Latin 87 Revival 88 Celtic 89 Bluegrass 90 Avantgarde 91 {Gothic Rock} \
			92 {Progressive Rock} 93 {Psychedelic Rock} 94 {Symphonic Rock} 95 {Slow Rock} \
			96 {Big Band} 97 Chorus 98 {Easy Listening} 99 Acoustic 100 Humour 101 Speech \
			102 Chanson 103 Opera 104 {Chamber Music} 105 Sonata 106 Symphony 107 {Booty Brass} \
			108 Primus 109 {Porn Groove} 110 Satire 111 {Slow Jam} 112 Club 113 Tango 114 Samba \
			115 Folklore 116 Ballad 117 {Poweer Ballad} 118 {Rhytmic Soul} 119 Freestyle \
			120 Duet 121 {Punk Rock} 122 {Drum Solo} 123 {A Capela} 124 Euro-House 125 {Dance Hall} \
			126 Goa 127 {Drum & Bass} 128 Club-House 129 Hardcore 130 Terror 131 Indie 132 BritPop \
			133 Negerpunk 134 {Polsk Punk} 135 Beat 136 {Christian Gangsta Rap} 137 {Heavy Metal} \
			138 {Black Metal} 139 Crossover 140 {Contemporary Christian} 141 {Christian Rock} \
			142 Merengue 143 Salsa 144 {Trash Metal} 145 Anime 146 JPop 147 SynthPop"

	# Grab the last 125 bytes of the mp3 (3 bytes ignored for "TAG", totaling 128 bytes)
	set fin [open $currFile r]
	seek $fin -128 end
	if {[read $fin 3] != "TAG"} {
		# The id3 tag does not exist correctly.
		return ""
	}
	set songTitle [read $fin 30]
	set artist [read $fin 30]
	set album [read $fin 30]
	set year [read $fin 4]
	set comment [read $fin 30]
	set genre [read $fin 1]
	close $fin
	# Right now $genre is in binary (not 1's and 0's, but machine-language, unreadable binary), so convert to binary numbers of 1's and 0's.
	binary scan $genre B* genre
	# Now convert binary to decimal.
	set myGenreNum [vcheck:binaryToDecimal $genre]
	# Finally look up the genre id to find what it pertains to (Ska, Rap, etc.)
	if {[catch {set genre $mp3Genre($myGenreNum)}]} {
		set genre "Unknown"
	}
	# Return a list and filter the strings with bad chars mysql will invoke an error with, and trailing control characters.
	return "\{[vcheck:filterString $songTitle]\} \{[vcheck:filterString $artist]\} \{[vcheck:filterString $album]\} \{$year\} \{[vcheck:filterString $comment]\} \{$genre\}"
}


proc vcheck:mp3:getHeader {} { global currFile
	set mpeg ""
	set fin [open $currFile r]
	set mp3Header [read $fin 4]
	close $fin
	# From machine binary to 1's and 0's (to test which bits are set).
	binary scan $mp3Header B* mp3Header
	if {[string index $mp3Header 12] == 1} {
		set mpeg 1
	} else {
		set mpeg 2
	}

	set layer ""
	#00 {set layer "none"}
	#changed to numerical for database datatype simplification
	switch -exact -- [string range $mp3Header 13 14] {
		11 {set layer 1}
		01 {set layer 3}
		10 {set layer 2}
		00 {set layer 0}
		default {set layer -1}
	}

	set protection ""
	if {[string index $mp3Header 15] == 1} {
		set protection 1
	} else {
		set protection 0
	}

	set bitrate ""
	#0000 {set bitrate "NULL"}
	#1111 {set bitrate "NULL"}
	#changed for database datatype simplification
	switch -exact -- [string range $mp3Header 16 19] {
		0000 {set bitrate 0}
		0001 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 32 32 32 8]}
		0010 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 64 48 40 16]}
		0011 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 96 56 48 24]}
		0100 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 128 64 56 32]}
		0101 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 160 80 64 64]}
		0110 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 192 96 80 80]}
		0111 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 224 112 96 56]}
		1000 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 256 128 112 64]}
		1001 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 288 160 128 128]}
		1010 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 320 192 160 160]}
		1011 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 352 224 192 112]}
		1100 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 384 256 224 128]}
		1101 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 416 320 256 256]}
		1110 {set bitrate [vcheck:mp3:findBitrate $mpeg $layer 448 384 320 320]}
		1111 {set bitrate 0}
		default {set bitrate -1}
	}

	set frequency ""
	#11 {set frequency "NULL"}
	#for database datatype simplification
	switch -exact -- [string range $mp3Header 20 21] {
		00 {
			if {$mpeg == 1} {
				set frequency 44100
			} else {
				set frequency 22050
			}
		}

		01 {
			if {$mpeg == 1} {
				set frequency 48000
			} else {
				set frequency 24000
			}
		}

		10 {
			if {$mpeg == 1} {
				set frequency 32000
			} else {
				set frequency 16000
			}
		}

		11 {set frequency 0}
		default {set frequency -1}
	}

	set padding [string index $mp3Header 22]
	set private [string index $mp3Header 23]

	set mode ""
	switch -exact -- [string range $mp3Header 24 25] {
		00 {set mode "stereo"}
		01 {set mode "joint stereo"}
		10 {set mode "dual channel"}
		11 {set mode "mono"}
		default {set mode -1}
	}

	set modeExtension [string range $mp3Header 26 27]
	set copyright [string index $mp3Header 28]
	set home [string index $mp3Header 29]

	set emphasis ""
	switch -exact -- [string range $mp3Header 30 31] {
		00 {set emphasis "none"}
		01 {set emphasis "50/15ms"}
		10 {set emphasis "NULL"}
		11 {set emphasis "CCITT j.17"}
		default {set emphasis -1}
	}

	set frames ""
	if {$padding == 1} {
		set frames [expr (144 * (($bitrate * 1000) / $frequency)) + 1]
	} else {
		set frames [expr 144 * (($bitrate * 1000) / $frequency)]
	}
	set seconds [expr [file size $currFile] / (($bitrate * 1000) / 8)]

	# We don't want to use all of the gathered information for now...just these.
	return "\{$mpeg\} \{$layer\} \{$bitrate\} \{$frequency\} \{$mode\}"
}


proc vcheck:stringToList {myString maxChars delimeter} {
	# Takes in a string and decimal, returns a list with each element <= $maxChars chars.  The elements are based on the parsing, which, in turn, is based on the delimiter (such as a whitespace).
	# Break the string into a list based on the provided delimiter.
	set myString [split $myString $delimeter]
	set myStringLen [llength $myString]
	# If it's only one element large, we have no choice but to send it out (we won't truncate).
	if {$myStringLen == 1} {
		return $myString
	}
	set a 0
	set myStringSpliced ""
	while {$a < $myStringLen} {
		# Start rebuilding the list (so put element, delimiter, element, delimiter...).
		set newElement "[lindex $myString $a]$delimeter"
		# Can we fit in the next element (it may not exceed $maxChars).  Note where $a is being incremented.
		while {([expr [string length $newElement] + [string length [set nextStringElement [lindex $myString [set a [expr $a + 1]]]]]] <= $maxChars) && ($nextStringElement != "")} {
			# Yes we can, so let's add to it until we can't, in which we have our "new" element (named spliced).
			append newElement "$nextStringElement$delimeter"
		}
		lappend myStringSpliced [string range $newElement 0 end-1]
	}
	return $myStringSpliced
}


proc vcheck:getReleaseName {currPath} { global subDirExp
	# If a release has cd1, cd2, or anything thereof, then this will return the actual "base" release's name.  /site/apps/my.cat/cd1 = "my.cat"
	set myRelease [file tail $currPath]
	set temp [string tolower $myRelease]
	foreach myRegexp $subDirExp {
		if {[string match $myRegexp $temp]} {
			return [file dirname $currPath]
		}
	}
	return $currPath
}


proc vcheck:completion:updateProgressBar {missingFilesAmount} { global argv
	catch {file delete -force -- "./[lindex $argv 20]"}
	foreach missingDir [glob -nocomplain "*_MISSING.*.FILES_*"] {
		file delete -force -- $missingDir
	}
	if {$missingFilesAmount == 0} {
		file mkdir "[lindex $argv 20]"
	} else {
		file mkdir "./\[_MISSING.$missingFilesAmount.FILES_\]"
	}
	return 0
}


proc vcheck:getDuration {secs} {
	set hrs 0
	set mins 0
	set secs [format "%.0f" $secs]
	if {$secs >= 3600} {
		for {set hrs 1} {$secs >= 3600} {incr hrs} {
			incr secs -3600
		}
	}
	if {$secs >= 60} {
		for {set mins 1} {$secs >= 60} {incr mins} {
			incr secs -60
		}
	}
	if {$hrs != 0} {
		return "${hrs}hrs ${mins}mins ${secs}secs"
	} elseif {$mins != 0} {
		return "${mins}mins ${secs}secs"
	}
	return "${secs}secs"
}


proc vcheck:getSafePath {path} { global argv env
	# Are we using OpenFTPD?
	if {[lindex $argv 21] == 2} {
		# Strip out the standardDir for openftpd.
		return [string range $path [expr [string length $env(F4A_STD)] - 1] end]
	}
	# Nothing special for glftpd. (you could strip out the first dir, usually "/site", but not guaranteed, could be "/")
	return $path
}


proc vcheck:db:updateIncompleteReleases {isFirstFile missingFiles} { global mysql currFile rid
	if {$isFirstFile != 1} {
		# Since this isn't the first file of the release, delete its entry from the incomplete table.
		sql exec $mysql(conn) "DELETE FROM IncompleteReleases WHERE releaseId=$rid AND fileName='$currFile';"
	} else {
		# This is the "first" file of the release.
		# Ensure no duplicate entries (in case the files were deleted and are being re-uploaded from file #1).
		sql exec $mysql(conn) "DELETE FROM IncompleteReleases WHERE releaseId=$rid;"
		# Since this *is* the first file of the release, add all missing files to the database.
		foreach myFile $missingFiles {
			sql exec $mysql(conn) "INSERT INTO IncompleteReleases VALUES ($rid, '$myFile');"
		}
	}
}



proc vcheck:race:getUserRacers {} { global rid
	# Query the user's current stats for this release (includes the file currently being processed).
	vcheck:db:query "SELECT SUM(fileSize) AS size, userName, groupName, COUNT(*) AS files, AVG(averageSpeed), SUM((fileSize/1024.0)/averageSpeed) AS seconds FROM ReleasesDetails WHERE releaseId=$rid GROUP BY userName ORDER BY size DESC;"
	if {[string index $rows 0] == "\{"} {
		# We can straight copy it because the query returned multiple users for us and has an outter brace: {{user1} {user2}}
		return $rows
	}
	# Does not have outer brace, so list-append to create it.  Only one user was returned with stats like: user1 (where user1 is a list itself of {bytes user group files speed seconds}).
	return "\{$rows\}"
}


proc vcheck:race:getGroupRacers {} { global rid
	vcheck:db:query "SELECT SUM(fileSize) AS size, groupName, COUNT(*) AS files, AVG(averageSpeed), SUM((fileSize/1024.0)/averageSpeed) AS seconds FROM ReleasesDetails WHERE releaseId=$rid GROUP BY groupName ORDER BY size DESC;"
	if {[string index $rows 0] == "\{"} {
		return $rows
	}
	return "\{$rows\}"
}


proc vcheck:race:isNewUserRacer {} { global rid currUser
	# Check to see if the user and group were previously in the race
	vcheck:db:query "SELECT releaseId FROM ReleasesDetails WHERE releaseId=$rid AND userName='$currUser' LIMIT 1;"
	# We basically want to know if this variable is null or has some actual value, and releaseId can be queried the quickest.
	if {$rows == ""} {
		return 1
	}
	return 0
}


proc vcheck:race:isNewGroupRacer {} { global rid currGroup
	vcheck:db:query "SELECT releaseId FROM ReleasesDetails WHERE releaseId=$rid AND groupName='$currGroup' LIMIT 1;"
	if {$rows == ""} {
		return 1
	}
	return 0
}


proc vcheck:race:isNewUserLeader {userRacers oldUserLeader} { global currUser
	# Check if the leading user racer is the person uploading this file, and was or was not previously leading.
	if {([lindex [lindex $userRacers 0] 1] == $currUser) && ($oldUserLeader != $currUser)} {
		# The user uploading just became a the new winner (leader) in the race because of this file.  Set to true.
		return 1
	}
	# The $currUser is either not leading or was already leading previous to this file upload.
	return 0
}


proc vcheck:race:isNewGroupLeader {groupRacers oldGroupLeader} { global currGroup
	if {([lindex [lindex $groupRacers 0] 1] == $currGroup) && ($oldGroupLeader != $currGroup)} {
		# A new group is in the lead.
		return 1
	}
	return 0
}


proc vcheck:race:getUserTag {} { global argv env
	# Is this glftpd?
	if {[lindex $argv 21] == 1} {
		return $env(TAGLINE)
	}
	# OpenFTPD.
	# The following will be available soon.
	#return $env(F4A_USERINFO)
}


proc vcheck:race:getGroupTag {group} { global argv
	# Using glftpd?
	if {[lindex $argv 21] == 1} {
		# Grab user's group's tag from the group file.
		set fin [open [lindex $argv 5] r]
		while {![eof $fin]} {
			set temp [split [gets $fin] ":"]
			if {[lindex $temp 0] == $group} {
				close $fin
				return [lindex $temp 1]
			}
		}
		close $fin
		# Could not find it?
		return ""
	}
	# OpenFTPD.
	# Available soon.
	#return $env(F4A_GROUPINFO)
}


proc vcheck:mp3:makeSortedSymlinks {genre} { global argv currPath
	# $myRelease is the full path to the actual release, with CD1, CD2..etc. all ignored (cleaned).  Think of it as a renamed $currPath.
	set myRelease ""
	# The tail, then, would be the actual release name.
	set myReleaseTail ""

	# Make genre symlink?
	if {[file isdirectory [lindex $argv 24]]} {
		set myRelease [vcheck:getReleaseName $currPath]
		set myReleaseTail [file tail $myRelease]
		set linkExists ""
		# $linkExists will be null if it does't exist, or "link" if it does.  The null case will come with an error, so we need to catch it.
		catch {set linkExists [file type "[lindex $argv 24]/$genre/$myReleaseTail"]}
		if {$linkExists == ""} {
			if {![file isdirectory "[lindex $argv 24]/$genre"]} {
				file mkdir "[lindex $argv 24]/$genre"
			}
			exec -- ln -s $myRelease "[lindex $argv 24]/$genre/$myReleaseTail" 2>@ stderr
		}
	}

	# Make alphanumeric symlink?
	if {[file isdirectory [lindex $argv 25]]} {
		if {$myRelease == ""} {
			set myRelease [vcheck:getReleaseName $currPath]
			set myReleaseTail [file tail $myRelease]
		}
		set alphaNumericDir [string toupper [string index $myReleaseTail 0]]
		set linkExists ""
		catch {set linkExists [file type "[lindex $argv 25]/$alphaNumericDir/$myReleaseTail"]}
		if {$linkExists == ""} {
			if {![file isdirectory "[lindex $argv 25]/$alphaNumericDir"]} {
				file mkdir "[lindex $argv 25]/$alphaNumericDir"
			}
			exec -- ln -s $myRelease "[lindex $argv 25]/$alphaNumericDir/$myReleaseTail" 2>@ stderr
		}
		unset alphaNumericDir
	}

	# Make group symlink?
	if {[file isdirectory [lindex $argv 26]]} {
		if {$myRelease == ""} {
			set myRelease [vcheck:getReleaseName $currPath]
			set myReleaseTail [file tail $myRelease]
		}
		set myReleaseSplit [split $myRelease "-"]
		if {[llength $myReleaseSplit] != 1} {
			set groupDir [string toupper [lindex $myReleaseSplit end]]
			set linkExists ""
			catch {set linkExists [file type "[lindex $argv 26]/$groupDir/$myReleaseTail"]}
			if {$linkExists == ""} {
				if {![file isdirectory "[lindex $argv 26]/$groupDir"]} {
					file mkdir "[lindex $argv 26]/$groupDir"
				}
				exec -- ln -s $myRelease "[lindex $argv 26]/$groupDir/$myReleaseTail" 2>@ stderr
			}
		}
	}
}


proc vcheck:race:makeIncompleteSymlinks {missingFilesAmount} { global argv currPath subDirExp
	# Now the symlink (for incomplete releases)
	if {[file isdirectory [lindex $argv 8]]} {
		set releaseLink "[lindex $argv 8]/[file tail $currPath]"
		set myCurrDir [string tolower [file tail $currPath]]
		foreach myRegexp $subDirExp {
			if {[string match $myRegexp $myCurrDir]} {
				# The current path is in a subdir of the release (cd1, et al) so let's include the release and subdir names in the symlink file.
				set myPathSplit [split $currPath "/"]
				set releaseLink "[lindex $argv 8]/[lindex $myPathSplit end-1]_[lindex $myPathSplit end]"
				break
			}
		}
		if {$missingFilesAmount != 0} {
			set linkExists ""
			catch {set linkExists [file type $releaseLink]}
			if {$linkExists == ""} {
				exec -- ln -s $currPath $releaseLink 2>@ stderr
			}
		} else {
			# The release is done; delete the incomplete symlink.
			file delete -force -- $releaseLink
		}
	}
}


proc vcheck:race:deleteMissingFiles {} { global argv currPath currFile
	# Do we need to delete a 0byte -missing file?
	if {[vcheck:matchPath [lindex $argv 27]]} {
		if {[file exist "$currPath/$currFile[lindex $argv 29]"]} {
			file delete -force -- "$currPath/$currFile[lindex $argv 29]"
		}
	}
}


proc vcheck:updateRace {currPath fileType completedFiles requiredFiles showCompletion extras newRequiredFiles} { global revision currFile argv env mysql currTime rid timeStampFile currUser currGroup currSpeed
	set missingFiles ""
	# If it's the first file, then we already we the requiredFiles per individual file processing (rar/mp3 or zip).  Refer to the caller procedure.
	if {([llength $completedFiles] == 1) || ($newRequiredFiles)} {
		# Simple equation really: requiredFiles - completedFiles = missingFiles. :-)
		set missingFiles [vcheck:completion:getMissingFilesCompare $requiredFiles $completedFiles]
		# Since it's first file uploaded, create records for all missingFiles in IncompleteReleases table.  Sending 1=this is the first file, 0=this is not the first file.
		vcheck:db:updateIncompleteReleases 1 $missingFiles
		# Create 0byte -missing files?
		if {[vcheck:matchPath [lindex $argv 27]]} {
			foreach myFile $missingFiles {
				set fout [open "$myFile[lindex $argv 29]" w]
				close $fout
			}
		}
	} else {
		# This isn't the first file, so let's refer to our database for all the goodies (rather than trying to parse sfv again, or figure the zip's filename incrementation).
		# Update the IncompletesReleases table by deleting the currFile from it.
		vcheck:db:updateIncompleteReleases 0 $missingFiles
		# Get missing files list even though we don't know what is required yet (because we have it in the database).
		set missingFiles [vcheck:completion:getMissingFilesDb]
		# Simple equation really: missingFiles + completedFiles = requiredFiles. :-)
		set requiredFiles [vcheck:completion:getRequiredFilesCombine $completedFiles $missingFiles]
	}

	# Verify that the release is large enough to continue for race info.
	if {[llength $requiredFiles] < [lindex $argv 18]} {
		vcheck:db:safeExit 0
	}


	#START main database updates
	# Is the user and group just entering the race? 1=yes, 0=no.
	set newUserRacer [vcheck:race:isNewUserRacer]
	set newGroupRacer [vcheck:race:isNewGroupRacer]

	# Keep the old winning user and group names in variables to later see (after this $currFile has been entered) if there is a *new* user/group that is winning (then post to logs).
	vcheck:db:query "SELECT userName, SUM(fileSize) AS size FROM ReleasesDetails WHERE releaseId=$rid GROUP BY userName ORDER BY size DESC LIMIT 1;"
	set oldUserLeader [lindex $rows 0]
	vcheck:db:query "SELECT groupName, SUM(fileSize) AS size FROM ReleasesDetails WHERE releaseId=$rid GROUP BY groupName ORDER BY size DESC LIMIT 1;"
	set oldGroupLeader [lindex $rows 0]

	#START main database entries
	# Add the current file being processed to the database (we are already assured the file is not in the ReleasesDetails table when $completedFiles was made).
	sql exec $mysql(conn) "INSERT INTO ReleasesDetails VALUES ($rid, '$currFile', [file size $currFile], $currSpeed, '$currUser', '$currGroup', '[clock format $currTime -format "%Y-%m-%d %H:%M:%S"]');"

	# Grab/set a few totals for later use.
	vcheck:db:query "SELECT SUM(fileSize), MIN(timeStamp), MAX(timeStamp) FROM ReleasesDetails WHERE releaseId=$rid GROUP BY releaseId;"
	set releaseSize [lindex $rows 0]
	set raceDuration [expr [clock scan [lindex $rows 2]] - [clock scan [lindex $rows 1]]]
	set releaseAvgSpeed ""
	# We'll use this several times later, so let's cache it.
	set missingFilesAmount [llength $missingFiles]

	# Create a list of the users' stats.
	set userRacers [vcheck:race:getUserRacers]
	set groupRacers [vcheck:race:getGroupRacers]

	# 1=is a new user/group leader, 0=not.
	set newUserLeader [vcheck:race:isNewUserLeader $userRacers $oldUserLeader]
	set newGroupLeader [vcheck:race:isNewGroupLeader $groupRacers $oldGroupLeader]
	#END main database updates


	# For easier/faster access to the winners' info.
	set uwon [lindex $userRacers 0]
	set gwon [lindex $groupRacers 0]

	# Get the taglines for the winners.
	set uwonTag [vcheck:race:getUserTag]
	# Send in the group name also.
	set gwonTag [vcheck:race:getGroupTag [lindex $gwon 1]]


	#START all the mp3 stuff
	set mpeg ""; set layer ""; set bitrate ""; set frequency ""; set mode ""
	set songTitle ""; set artist ""; set album ""; set year ""; set comment ""; set genre ""
	# mp3 id3/header enabled?
	if {($fileType == "*.mp3") && ([vcheck:matchPath [lindex $argv 16]])} {
		# Query the header/id3 info.
		vcheck:db:query "SELECT mpeg, layer, bitrate, frequency, mode, title, artist, album, year, comment, genre FROM Mp3Releases WHERE releaseId=$rid;"
		if {$rows == ""} {
			# The info hasn't been entered yet, so let's generate it and put it in the database (header + id3).
			set temp [vcheck:mp3:getHeader]
			set mpeg [lindex $temp 0]
			set layer [lindex $temp 1]
			set bitrate [lindex $temp 2]
			set frequency [lindex $temp 3]
			set mode [lindex $temp 4]
			set temp [vcheck:mp3:getId3]
			if {$temp != ""} {
				set songTitle [lindex $temp 0]
				set artist [lindex $temp 1]
				set album [lindex $temp 2]
				set year [lindex $temp 3]
				set comment [lindex $temp 4]
				set genre [lindex $temp 5]
				# As you see, id3 tag must exist or even the header info is not inserted to the database (I could split the tables, but, nah :)).
				sql exec $mysql(conn) "INSERT INTO Mp3Releases VALUES ($rid, $mpeg, $layer, $bitrate, $frequency, '$mode', '$songTitle', '$artist', '$album', '$year', '$comment', '$genre');"
			}
		} else {
			set mpeg [lindex $rows 0]
			set layer [lindex $rows 1]
			set bitrate [lindex $rows 2]
			set frequency [lindex $rows 3]
			set mode [lindex $rows 4]
			set songTitle [lindex $rows 5]
			set artist [lindex $rows 6]
			set album [lindex $rows 7]
			set year [lindex $rows 8]
			set comment [lindex $rows 9]
			set genre [lindex $rows 10]
		}

		# Is the release done (note: still in mp3 branch)?
		if {$missingFilesAmount == 0} {
			# Make a nifty mp3 id3/header dir?
			if {[vcheck:matchPath [lindex $argv 23]]} {
				catch {file mkdir "\[$year - $genre - ${bitrate}kbps - $mode\]"}
			}

			# Make all the symlinks.
			vcheck:mp3:makeSortedSymlinks $genre
		}
	}
	#END all the mp3 stuff


	set releasePercent ""
	if {$showCompletion} {
	  	set releasePercent [expr (100.0 * [llength $completedFiles]) / [llength $requiredFiles]]
		# Update the progress dir?
	  	if {[lindex $argv 20] != ""} {
	  		vcheck:completion:updateProgressBar $missingFilesAmount
	  	}

		# Make symlinks to incomplete releases or delete it if done.
		vcheck:race:makeIncompleteSymlinks $missingFilesAmount

		if {$missingFilesAmount == 0} {
			# Delete the cached timestamp since there are no more missing files.  Once the timestamp is gone, as far as the database is concerned, this is a new release (if files are re-uploaded).
			if {![vcheck:matchPath [lindex $argv 3]]} {
				catch {file delete -force -- $timeStampFile}
			}

			# And kill the cached .file_id.diz if we're done with this zip release (note, this cache may not pertain to completion, so we treat it individually).
			if {$fileType == "*.zip"} {
				catch {file delete -force -- .file_id.diz}
			}
		}

		# Delete our 0byte -missing file if needed.
		vcheck:race:deleteMissingFiles
	}

	# Party time...write the .message :-)
	if {[vcheck:matchPath [lindex $argv 17]]} {
		if {$fileType == "*.zip"} {
			# Let's append in case a diz header exists in it already.
			set raceinfo [open .message a]
			puts $raceinfo "\n\n\n"
		} else {
			set raceinfo [open .message w]
		}

		# Send out a quick intro newline.
		puts ""

		puts $raceinfo "                         --< vCheck Revision: $revision >--"
		puts $raceinfo ""
		puts $raceinfo "!-========(   Experience the moment with peace, love, and without anxiety   )========-!"
		puts $raceinfo "! Dir         ->  [format "%-67s" "[vcheck:getSafePath $currPath]"] !"
		if {($fileType == "*.\[r0-9\]\[a0-9\]\[r0-9\]") || ($fileType == "*.mp3")} {
			if {$extras != ""} {
				puts $raceinfo "! SFV date    ->  [format "%-67s" "[clock format [lindex $extras 0] -format "%B %d, %Y"]"] !"
			}
		}
		puts $raceinfo "! Updated     ->  [format "%-67s" "[clock format $currTime -format "%I:%M%p %Z"]"] !"
		if {$raceDuration > 0} {
			puts $raceinfo "! [format "%-83s" "Race lasted ->  [vcheck:getDuration $raceDuration]"] !"
		}
		if {$showCompletion} {
			puts $raceinfo "! Status      ->  [format "%-67s" "[llength $completedFiles] of [llength $requiredFiles] files uploaded ( [format "%6.2f%%" "$releasePercent"] )"] !"
			puts "-<   [llength $completedFiles] of [llength $requiredFiles] files uploaded: [format "%.0f" "$releasePercent"]%   >-"
			if {$missingFilesAmount > 0} {
				puts $raceinfo "! [format "%-83s" "Missing files:"] !"
				foreach temp [vcheck:stringToList $missingFiles 78 " "] {
					puts $raceinfo "!   -> [format "%-78s" "$temp"] !"
				}
			}
		}
		puts $raceinfo "`-===================================================================================='"

		if {$mpeg != "" || $songTitle != ""} {
			puts $raceinfo ""
			puts $raceinfo "...::(  Muzik!  )::..."
			puts $raceinfo ".-===================================================================================-."
			if {$mpeg != ""} {
				puts $raceinfo "| [format "%83s" "MPEG-$mpeg, Layer-$layer, ${bitrate}kb/s, ${frequency}Hz    "] |"
			}
			if {$songTitle != ""} {
				puts $raceinfo "| [format "%-83s" "Title      : $songTitle"] |"
				puts $raceinfo "| [format "%-83s" "Artist     : $artist"] |"
				puts $raceinfo "| [format "%-83s" "Album      : $album - $year"] |"
				puts $raceinfo "| [format "%-83s" "Comment    : $comment"] |"
				puts $raceinfo "| [format "%-83s" "Genre      : $genre"] |"
			}
			puts $raceinfo "`-===================================================================================='"
		}

		puts $raceinfo ""
		puts $raceinfo "...::(  The Cows!  )::..."
		puts $raceinfo ".-===================================================================================-."
		puts $raceinfo "| User Winner:  [format "%-69s" "[lindex $uwon 1] / $uwonTag"] |"
		puts $raceinfo "|-===================================================================================-|"
		puts $raceinfo "| ## /       User@Group       /       Size       /  Files  /  Avg Speed  / Percentage |"
		puts $raceinfo "|||||||"

		puts ""
		puts "::\[ user ranks \]::"

		# userRacers = {bytes user group files speed seconds}
		set num 0;  set totalAvgSpeed 0
		foreach {bytes user group files speed seconds} [join $userRacers] {
			incr num
			set totalAvgSpeed [expr $totalAvgSpeed + $speed]
			puts $raceinfo "| [format "%-2d" $num] |[format "%23s" "$user@$group"] |[format "%15d" [expr $bytes / 1024]]KB |[format "%8d" $files] |[format "%8.0f" $speed]kB/s |[format "%10.2f" [expr (100.0 * $bytes) / $releaseSize]]% |"
			puts "<[format "%-2d" $num]> [format "%23s" "$user@$group"] [format "%8d" [expr $bytes / 1048576]]MB [format "%7d" $files] files [format "%8.0f" $speed]kB/s"
		}
		set totalAvgSpeed [expr $totalAvgSpeed / [llength $userRacers]]

		puts $raceinfo "|____|________________________|__________________|_________|_____________|____________|"
		puts $raceinfo "| [format "%-2d" [llength $userRacers]] |[format "%23s" ""] |[format "%15d" [expr $releaseSize / 1024]]KB |[format "%8d" [llength $completedFiles]] |[format "%8.0f" $totalAvgSpeed]kB/s |[format "%10.2f" 100]% |"
		puts $raceinfo "`-===================================================================================='"
		set releaseAvgSpeed $totalAvgSpeed

		puts $raceinfo ""
		puts $raceinfo "...::(  The Barns!  )::..."
		puts $raceinfo ".-===================================================================================-."
		puts $raceinfo "| Group Winner:  [format "%-68s" "[lindex $gwon 1] / $gwonTag"] |"
		puts $raceinfo "|-===================================================================================-|"
		puts $raceinfo "| ## /         Group          /       Size       /  Files  /  Avg Speed  / Percentage |"
		puts $raceinfo "|||||||"

		puts ""
		puts "::\[ group ranks \]::"

		#groupRacers = {bytes group files speed seconds}
		set num 0;  set totalAvgSpeed 0
		foreach {bytes group files speed seconds} [join $groupRacers] {
			incr num
			set totalAvgSpeed [expr $totalAvgSpeed + $speed]
			puts $raceinfo "| [format "%-2d" $num] |[format "%23s" $group] |[format "%15d" [expr $bytes / 1024]]KB |[format "%8d" $files] |[format "%8.0f" $speed]kB/s |[format "%10.2f" [expr (100.0 * $bytes) / $releaseSize]]% |"
			puts "<[format "%-2d" $num]> [format "%23s" $group] [format "%8d" [expr $bytes / 1048576]]MB [format "%7d" $files] files [format "%8.0f" $speed]kB/s"
		}
		set totalAvgSpeed [expr $totalAvgSpeed / [llength $groupRacers]]

		puts $raceinfo "|____|________________________|__________________|_________|_____________|____________|"
		puts $raceinfo "| [format "%-2d" [llength $groupRacers]] |[format "%23s" ""] |[format "%15d" [expr $releaseSize / 1024]]KB |[format "%8d" [llength $completedFiles]] |[format "%8.0f" $totalAvgSpeed]kB/s |[format "%10.2f" 100]% |"
		puts $raceinfo "`-===================================================================================='"

		puts "\n"

		close $raceinfo
	}

	# Log it babie!
	if {$newUserRacer} {
		if {[llength $userRacers] == 1} {
			puts "Oooohhh...starting the race like a true leader.  3ye f34r j00 $currUser ;/"
		} else {
			if {[vcheck:matchPath [lindex $argv 28]]} {
				# Source the sitebot so we can  use its procedures.
				vcheck:siteBot NEW_URACER $userRacers $groupRacers $requiredFiles $completedFiles $missingFiles $raceDuration $releaseAvgSpeed $releaseSize $uwonTag $gwonTag "\{$genre\} $year $bitrate \{$mode\} \{$album\} \{$artist\} \{$songTitle\} \{$comment\} $mpeg $layer $frequency"
			}
			puts "So you're new to this place eh?  Well buckle-up and let's riiiiiiide! ;D"
		}
	}
	if {$newGroupRacer} {
		if {[llength $groupRacers] == 1} {
			puts "May $currGroup rule this release like a 5-on-1 gangbang!"
		} else {
			# Too much spamming on irc, but if you want it, enable and implement it yourself.
			#vcheck:siteBot NEW_GRACER $userRacers $groupRacers $requiredFiles $completedFiles $missingFiles $raceDuration $releaseAvgSpeed $releaseSize $uwonTag $gwonTag {$genre $year $bitrate $mode $album $artist $songTitle $comment $mpeg $layer $frequency}
			puts "Teehee...you think $currGroup is gonna win? :P  Prove me wrong!"
		}
	}
	if {$newUserLeader} {
		if {[llength $userRacers] != 1} {
			#vcheck:siteBot NEW_ULEADER $userRacers $groupRacers $requiredFiles $completedFiles $missingFiles $raceDuration $releaseAvgSpeed $releaseSize $uwonTag $gwonTag {$genre $year $bitrate $mode $album $artist $songTitle $comment $mpeg $layer $frequency}
			puts "Awww ya mah niggah, j00 just pulled ahead of $oldUserLeader...u iz in da lead f00! :D"
		}
	}
	if {$newGroupLeader} {
		if {[llength $groupRacers] != 1} {
			#vcheck:siteBot NEW_GLEADER $userRacers $groupRacers $requiredFiles $completedFiles $missingFiles $raceDuration $releaseAvgSpeed $releaseSize $uwonTag $gwonTag {$genre $year $bitrate $mode $album $artist $songTitle $comment $mpeg $layer $frequency}
			puts "B0W DOWN BABIE.  $currGroup pulled ahead of $oldGroupLeader.  day is in da fuckin hause!@#\$  y0 grp took the lead!  Show dem nigglets who tha gang bangerz are ;)"
		}
	}
	if {$missingFilesAmount == 0 && ($showCompletion)} {
		vcheck:siteBot RACE_COMPLETE $userRacers $groupRacers $requiredFiles $completedFiles $missingFiles $raceDuration $releaseAvgSpeed $releaseSize $uwonTag $gwonTag "\{$genre\} $year $bitrate \{$mode\} \{$album\} \{$artist\} \{$songTitle\} \{$comment\} $mpeg $layer $frequency"
		puts "yay, you finished the release.  relax, kick back, smoke a blunt, fuck some cows..."
	}
	puts ""
	return 0
}


proc vcheck:getSfvCrc {sfv} { global currFile
	set fin [open $sfv r]
	while {![eof $fin]} {
		set currLine [split [string tolower [gets $fin]] " "]
		# Is it our file?  No need to skip comments, this is faster.
		if {[lindex $currLine 0] == $currFile} {
			# Return the CRC.
			return [string toupper [lindex $currLine 1]]
		}
	}
	# File was not found in the sfv file?
	return ""
}


proc vcheck:processSfv:getSfv {} {
	set sfv [glob -nocomplain "*.sfv"]
	if {[llength $sfv] > 1} {
		puts "Warning: SFV: Multiple SFVs have been found; I'm choosing to use \"[lindex $sfv 0]\" out of [llength $sfv] total: $sfv"
		set sfv [lindex $sfv 0]
		vcheck:validateSfv [lindex $sfv 0]
		return [lindex $sfv 0]
	} elseif {[llength $sfv] == 1} {
		return $sfv
	} else {
		puts "Error: SFV: I couldn't find any SFVs.  I can't do anything until someone gives me one!"
		vcheck:db:safeExit 2
	}
}


proc vcheck:processSfv:getSfvDate {sfv} {
	set fin [open $sfv r]
	while {![eof $fin]} {
		set currLine [string tolower [gets $fin]]
		# The following are sfv-date formats that are supported (any others to add?).
		#; Generated by WIN-SFV32 v1.01a on 2000-01-23 at 02:48.41
		#; Generated by SFV32nix v1.2 on 2000-01-30 at 02:39.29
		#; Generated by WIN-SFV32 v1.5 on 1/6/2000 at 10:13:47 PM
		#; Generated by pdSFV v1.2 (WIN) on 2001-01-20 at 21:22.01
		if {[string match ";*generated by *sfv* * on * at *" $currLine]} {
			# Dumb pdsfv is costing me more lines of code. :-)
			if {[string index [lindex $currLine 5] 0] == "\("} {
				set currLine [lreplace $currLine 5 5]
			}
			set pkgDate [lindex $currLine 6]
			if {[string match "*-*-*" $pkgDate]} {
				set pkgDate [split $pkgDate "-"]
				# Change from YYYY-DD-MM to MM/DD/YYYY
				set pkgDate "[lindex $pkgDate 1]/[lindex $pkgDate 2]/[lindex $pkgDate 0]"
			}
			if {[catch {set pkgDate [clock scan $pkgDate]}]} {
				# I guess it wasn't a real date, but we matched the string and parsed it!  Use this area to debug for more sfv-date formats.
				puts "Warning: Could NOT determine sfv date from \"$pkgDate\".  Email this to moopowah@yahoo.com please!"
				set pkgDate ""
			} else {
				close $fin
				return $pkgDate
			}
		}
	}
	close $fin
	# Date was not found.
	return ""
}


proc vcheck:processSfv:getRequiredFiles {sfv fileType} {
	set requiredFiles ""
	set fin [open $sfv r]
	while {![eof $fin]} {
		set currLine [string tolower [gets $fin]]
		if {([string index $currLine 0] == ";") || (![string match $fileType [lindex $currLine 0]])} {
			# If it's a comment or not of a "valid" file type, skip it (this means jpgs, or any other funky files in an SFV will not be included in the vcheck:updateRace info)
			continue
		}
		if {[llength $currLine] != 2} {
			puts "Error: Bad SFV format: \"$sfv\""
			vcheck:db:safeExit 2
		}
		# It passed the test, add it on to the list.
		lappend requiredFiles [lindex $currLine 0]
	}
	close $fin
	return $requiredFiles
}


proc vcheck:completion:getMissingFilesDb {} { global rid
	# This isn't the first file of the release (it's cached in the database, so let's not parse the sfv again to get requiredFiles -> missingFiles!) :)
	vcheck:db:query "SELECT fileName FROM IncompleteReleases WHERE releaseId=$rid;"
	return $rows
}


proc vcheck:completion:getMissingFilesCompare {requiredFiles completedFiles} {
	# Build and return a list of missing files by comparing it to requiredFiles (which was obtained by parsing the sfv file).
	set missingFiles ""
	foreach myFile $requiredFiles {
		if {[lsearch $completedFiles $myFile] == -1} {
			# If the file is required and it's not in our $completedFiles list, then it's missing!
			lappend missingFiles $myFile
		}
	}
	return $missingFiles
}


proc vcheck:completion:getRequiredFilesCombine {completedFiles missingFiles} {
	set requiredFiles ""
	# Now build the required files list by combining missingFiles and completedFiles (thus not re-parsing the sfv).
	foreach myFile $completedFiles {
		lappend requiredFiles $myFile
	}
	foreach myFile $missingFiles {
		lappend requiredFiles $myFile
	}
	return $requiredFiles
}


proc vcheck:processSfv {fileType} { global currPath currFile currCrc argv
	# Grab the sfv file.
	set sfv [vcheck:processSfv:getSfv]

	# File integrity check, assumed bad.
	set goodFile 0
	if {[vcheck:matchPath [lindex $argv 6]]} {
		# Use the checksum util to get the file's crc
		set env(ASFV_FILE) $currFile
		set env(ASFV_DIR) $testPath
		# I'm using a modified version of aSFV v2.341 which only has two exit codes: 0=good, 1=bad.
		if {[catch {exec -- [lindex $argv 4]/asfv 2>@ stderr} err]} {
			puts $err
		} else {
			set goodFile 1
		}
		# Don't need this env vars anymore, so let's trash them.
		unset env(ASFV_FILE)
		unset env(ASFV_DIR)
	} elseif {[vcheck:matchPath [lindex $argv 7]]} {
		# Grab the crc from the sfv file.
		set checksum [vcheck:getSfvCrc $sfv]
		# The file's crc was read in to $checksum and daemon provided the crc.
		if {$currCrc == $checksum} {
			set goodFile 1
		}
	} else {
		# Files in this dir are OK no matter what.
		set goodFile 1
	}

	# Output screen to user and exit.
	if {!$goodFile} {
		puts "Failure: CRC: Your file has been denied."
		vcheck:db:safeExit 2
	}
	unset goodFile

	# If a denyPath, then bail out and simply be left integrity-checked (most likely a privpath).
	if {[vcheck:matchPath [lindex $argv 19]]} {
		vcheck:db:safeExit 0
	}


	#SFV COMPLETION ROUTINE -- START
	vcheck:db:connect
	set completedFiles [vcheck:completion:getCompletedFiles]

	# Get the sfv's date and $requiredFiles list.  The sfv is parsed several times because it promotes modular code.  We aren't *that* worried about speed.  This is a progrmmer's decision. :)
	set pkgDate [vcheck:processSfv:getSfvDate $sfv]

	# Load up our sitebot's config procedures.
	source "[lindex $argv 4]/vcheck/sitebot.conf"

	set requiredFiles ""
	if {[llength $completedFiles] == 1} {
		# This is the "first" file uploaded (or others were deleted and this was re-uploaded), so we don't know the files that are required from us; read the .sfv to figure it out (this way, we only parse the sfv *once* per release to found out what is required).
		set requiredFiles [vcheck:processSfv:getRequiredFiles $sfv $fileType]

		# Announce expected files stuff to irc?
		if {[vcheck:matchPath [lindex $argv 9]]} {
			#vcheck:siteBot EXPECTED_FILES $currPath [llength $requiredFiles] [file size $currFile]
		}
	}
	#SFV COMPLETION ROUTINE -- END


	vcheck:updateRace $currPath $fileType $completedFiles $requiredFiles 1 $pkgDate 0
}


proc vcheck:processZip:verifyBases {currentFiles baseFileName} {
	# Return list if first and last file have the same base structure, -1 if not
	foreach myFile $currentFiles {
		if {![string match "${baseFileName}??.zip" $myFile]} {
			return ""
		}
	}
	return $currentFiles
}


proc vcheck:processZip:getDiskCount {dizFile} {
	# It is assumed no releases exceed 99 disks (should this be changed?).
	set fin [open $dizFile r]
	set totalFiles -1
	while {![eof $fin]} {
		set currLine [string tolower [gets $fin]]
		# The following regexp line should match the following styles of diz disk counts:
		# [o1/o3], [o1/03], [o1/3], [xx/3], [x/3], [disk:o1/o3], [1/3], (o1/o3), (o1/03), <o1/o3>, <1/3>, etc...
		# The next regexp does: xx/o3, xx/03, xx/3 (anything broader than this (e.g., 3/3) would match the date as well and we can't risk that...group's fault for being overly elite). =]
		# If you have more diz styles, or if I bunched up too many formats in one line here, mail me!
		# I had to use `.(?:.|)' instead of `{1,2}' because I don't know how to use the bounds yet; please help me!
		if {[regexp {[\[\(\<](?:disk\:|).(?:.|)(?:\s\/\s|[\/\\\|])((?:\d|o|)\d)[\>\)\]]} $currLine myMatch totalFiles]
				|| [regexp {xx\/((?:\d|o|)\d)} $currLine myMatch totalFiles]} {
			# We got a match, let's bail.
			break
		}
	}
	close $fin
	if {([string index $totalFiles 0] == "o") || ([string index $totalFiles 0] == "0")} {
		return [string index $totalFiles 1]
	}
	return $totalFiles
}


proc vcheck:db:isMissingFiles {} { global rid
	# Return 1 if the releaseId is missing any files, else 0.
	vcheck:db:query "SELECT releaseId FROM IncompleteReleases WHERE releaseId=$rid LIMIT 1;"
	if {$rows == ""} {
		return 0
	}
	return 1
}



proc vcheck:processZip:getRequiredFiles {totalFiles baseFileName} { global currFile
	# We will support three styles of file naming (multi-disk releases):
	# 1) myfile01.zip, myfile02.zip ... myfile10.zip, myfile11.zip, ...
	# 2) myfile1.zip, myfile2.zip ... myfile10.zip, myfile11.zip, ...
	# 3) myfilea.zip, myfileb.zip ... myfilej.zip, myfilek.zip, ... (this can also have -a, -b, -c, ...)

	set requiredFiles ""
	# $baseFileName is the portion of the filename that does not deal with the incrementation.  By default, it is received as all but the last two characters BEFORE the ".zip"
	if {$totalFiles == 1} {
		return $currFile
	} elseif {[regexp {^.*(0)\d\.zip$} $currFile]} {
		# This is incremented with numbers.  Example includes style #1.
		set leadingZero "0"
		for {set a 1} {$a <= $totalFiles} {incr a} {
			lappend requiredFiles "$baseFileName$leadingZero$a.zip"
			if {$a == 9} {
				# Set it to null because the next time through the loop it'll look for: myfile010.zip (we don't need the leading zero once we hit 10, if it existed in the first place).
				set leadingZero ""
			}
		}
	} elseif {[regexp {^.*[1-9]\.zip$} $currFile]} {
		# This is incremented with numbers.  Example includes style #2.  However, if the release is ten or more disks, we will *ASSUME* it is style #1 (any rip experts care to consult with me?).
		# We'll assume the release is not over 9 files so we need to include the "missing" character two positions to the left of the period in ".zip" (i.e., the character before the numbering).
		# So if the file is "myfile51.zip" then $leadingZero is now "5"
		set leadingZero [string index $currFile end-5]
		if {$totalFiles >= 10} {
			set leadingZero "0"
		}
		for {set a 1} {$a <= $totalFiles} {incr a} {
			lappend requiredFiles "$baseFileName$leadingZero$a.zip"
			if {$a == 9} {
				# Set it to null because the next time through the loop it'll look for: myfile010.zip (we don't need the leading zero once we hit 10, if it existed in the first place).
				set leadingZero ""
			}
		}
	} elseif {[regexp {^.*\w\.zip$} $currFile]} {
		# This is incremented with letters.  Example is style #3 above.
		# The numbering is by letters yet it's over 26 files?  Something went wrong.
		if {$totalFiles > 26} {
			puts "Warning: file numbering: there are $totalFiles in this release yet the files are incremented by letters?  Report this release to moopowah@yahoo.com PLEASE."
			return ""
		}
		set alphabet "abcdefghijklmnopqrstuvxyz"
		# The base file is all but the last character *before* the suffix ".zip"
		set baseFileName [string range $currFile 0 end-5]
		for {set a 0} {$a < $totalFiles} {incr a} {
			lappend requiredFiles "$baseFileName[string index $alphabet $a].zip"
		}
	}
	return $requiredFiles
}


proc vcheck:processZip {fileType} { global currPath currFile argv currTime
	# File integrity
	if {[vcheck:matchPath [lindex $argv 10]]} {
		if {[catch {exec -- unzip -tqq $currFile 2>@ stderr} err] == 1} {
			puts $err
			puts "Failure: CRC: Your file has been denied."
			vcheck:db:safeExit 2
		}
	}

	# Zip isn't in the zipRoutine dir?  bail out to avoid completion clutter between iso/mp3 and zip!
	if {![vcheck:matchPath [lindex $argv 11]]} {
	  	vcheck:db:safeExit 0
	}

	set newDiz 0
	# Want file_id.diz in the .message header? etc.  Cache and parse for file amount later.
	if {[vcheck:matchPath [lindex $argv 13]]} {
		# file_id.diz
		if {(![file exist .file_id.diz]) || ([file size .file_id.diz] == 0)} {
			catch {file delete -force -- .file_id.diz}
			catch {exec -- unzip -Cpoqq $currFile file_id.diz > .file_id.diz 2>@ stderr}
			set newDiz 1
		}
		catch {file copy -force -- .file_id.diz .message}
	} else {
		# For zipCompletion later then.
		if {(![file exist .file_id.diz]) || ([file size .file_id.diz] == 0)} {
			catch {file delete -force -- .file_id.diz}
			catch {exec -- unzip -Cpoqq $currFile file_id.diz > .file_id.diz 2>@ stderr}
			# Delete it because we will later open it for appending (and if it already exists, it'll always append!)
			catch {file delete -force -- .message}
			set newDiz 1
		}
	}

	# Comment
	if {[lindex $argv 14] != ""} {
		catch {exec -- zip -z -q $currFile < "[lindex $argv 14]" 2>@ stderr}
	}
	# site.nfo
	if {[lindex $argv 15] != ""} {
		catch {exec -- zip -j -q $currFile "[lindex $argv 15]" 2>@ stderr}
	}

	# If a denyPaths then bail out and simply be left integrity-checked.
	if {[vcheck:matchPath [lindex $argv 19]]} {
		vcheck:db:safeExit 0
	}


	#ZIP COMPLETION ROUTINE -- START
	vcheck:db:connect
	set completedFiles [vcheck:completion:getCompletedFiles]

	# Load up our sitebot's config procedures.
	source "[lindex $argv 4]/vcheck/sitebot.conf"

	set requiredFiles ""
	# If it is the first file of the release, _or_ it is a new diz file (maybe the previous files' diz were bad!) then try to parse the diz [again].
	if {([llength $completedFiles] == 1) || ($newDiz)} {
		# Routine to find total files in the release (from diz)
		set totalFiles [vcheck:processZip:getDiskCount .file_id.diz]

		if {$totalFiles == -1} {
			puts "Warning: I can't determine how many disks are in this release, I'll have to skip completion routine (send email to moopowah@yahoo.com about this release please!)."
		  	vcheck:updateRace $currPath $fileType $completedFiles "" 0 "" 0
			vcheck:db:safeExit 0
		}

		# Verify that the release is large enough to continue.
		if {$totalFiles < [lindex $argv 18]} {
			vcheck:db:safeExit 0
		}

		# Announce expected files stuff to irc?
		if {[vcheck:matchPath [lindex $argv 9]]} {
			#vcheck:siteBot EXPECTED_FILES $currPath $totalFiles [file size $currFile]
		}

		# The base file is all but the last two character *before* the suffix ".zip"
		set baseFileName [string range $currFile 0 end-6]
		if {[vcheck:processZip:verifyBases $completedFiles $baseFileName] == ""} {
			puts "Warning: File structure: The disk names that are required can't be determined (inconsistent bases), but there are a total of $totalFiles disks.  Please report this release to moopowah@yahoo.com"
			vcheck:updateRace $currPath $fileType $completedFiles "" 0 "" 0
			vcheck:db:safeExit 0
		}

		# Run our fancy procedure to figure out the syntax of the file names for zip.
		set requiredFiles [vcheck:processZip:getRequiredFiles $totalFiles $baseFileName]
	} elseif {![vcheck:db:isMissingFiles]} {
		# Notice the !isMissingFiles.  This means that there should be at least *one* file from this releaseId in the IncompleteReleases table (and also means this is not the first file being uploaded since it says [llength $completedFiles]. and the latter would only be evaluated if the first was false for ||).  However, if no missing files are found, this implies that vCheck could not determine the totalFiles for this release last time.
		puts "Warning: vCheck could not determine the amount of disks in this release on the previous file's file_id.diz (maybe it was a bad diz, its format isn't supported, or the disk count isn't even in the diz) so the completion will be skipped.  Please email vman (moopowah@yahoo.com) about this particular release."
	  	vcheck:updateRace $currPath $fileType $completedFiles "" 0 "" 0
		vcheck:db:safeExit 0
	}
	#ZIP COMPLETION ROUTINE -- END


	vcheck:updateRace $currPath $fileType $completedFiles $requiredFiles 1 "" $newDiz
}


proc vcheck:init {} { global argv
	uplevel #0 {
		set revision "0.1.0a"
		set sfvMaxSize		50000			;# don't parse sfv files larger than this many bytes
		set mysql(conn)		""				;# mysql handle
		set mysql(db)		"vcheck"		;# mysql database name
		set rid				""				;# release id for database
		set timeStampFile	".racestart"	;# a timestamp file is needed to distinguish records with the same path in the database
		# Keep it lowercase, subdir names for symlinks
		set subDirExp {
			"cd\[0-9\]"
			"dis\[ck\]\[0-9\]"
			"part\[0-9\]"
		}
	}

	if {[llength $argv] != 30} {
		puts "vCheck v$revision   -   Copyright (c) 1999-2001, vman (moopowah@yahoo.com)"
		puts "Don't run this file by hand, run: vcheck.sh <file> <dir> \[crc\]"
		exit 2
	}
}


proc vcheck:main {} { global currFile currPath
	uplevel #0 {
		set currFile [lindex $argv 0]
		# Not using [lindex $argv 1] (second argument) for openftpd compatibility (any problems with glftpd? tell me!)
		set currPath [pwd]
		# The current time to use for all things (logging, timeStamp, etc.)
		set currTime [clock seconds]
		if {[lindex $argv 21] == 1} {
			# glFtpD
			set currUser $env(USER)
			set currGroup $env(GROUP)
			set currSpeed $env(SPEED)
			set currCrc [lindex $argv 2]
		} elseif {[lindex $argv 21] == 2} {
			# OpenFtpD
			set currUser $env(F4A_USER)
			set currGroup $env(F4A_GROUP)
			set currSpeed [expr [lindex $argv 1] / 1024]
			set currCrc [lindex $argv 2]
		} else {
			puts "What daemon are *you* trying to use? :-)"
			vcheck:safeExit 2
		}
	}

	if {[catch {cd $currPath} err]} {
		puts $err
		puts "-> Can't change dir to \"$currPath\".  Don't look now, but I think you just got nuked! :P"
		vcheck:db:safeExit 2
	}

	# If the file is a ".ok" file, then rename it.  Check proc for more info.
	vcheck:checkForOk

	switch -regexp -- $currFile {
		{^.*\.([rs\d]\d\d|rar)$} {			# RAR
			vcheck:processSfv "*.\[rs0-9\]\[a0-9\]\[r0-9\]"
		}
	
		{^.*\.mp3$} {						# MP3
			vcheck:processSfv "*.mp3"
		}
	
		{^.*\.zip$} {						# ZIP
			vcheck:processZip "*.zip"
		}
	
		{^.*\.([c\d]\d\d|ace)$} {			# ACE
			vcheck:processSfv "*.\[ac0-9\]\[c0-9\]\[e0-9\]"
		}
	}

	# Bye folks. :-)
	vcheck:db:safeExit 0
}


# Start main()
vcheck:init
vcheck:main




################################################################################
# EOF #
#######

